import { isFloat } from "./utils";

// The default excel delimiter is ;
const FIELD_DELIMITER = ";"

interface CSVHeader {
    id: string,
    title: string,
    transformFunc?: (columnValue: any) => string
}

export default class CSVGenerator {
    private readonly locale: string;
    private readonly headers: CSVHeader[];

    constructor(headers: CSVHeader[], locale: string) {
        if (headers === undefined || headers.length === 0) {
            throw new Error("Provide at least one header to generate the CSV file")
        }
        this.headers = headers;
        this.locale = locale;
    }

    private getHeaders(visibleHeaders?: string[]) {
        let headersString = '';
        for (let header of this.headers) {
            if (visibleHeaders !== undefined) {
                if (visibleHeaders.includes(header.id)) {
                    headersString += header.title + ";";
                }
            } else {
                headersString += header.title + ";";
            }
        }
        return headersString + "\n";
    }

    private searchPropertyValues(data: { [key: string]: any }, column: string) {

        const objectValues = [];

        const getValues = (data: { [key: string]: any }, column: string) => {
            if (data[column] !== undefined) {
                objectValues.push(data[column]);
            } else {
                const splittedColumn = column.split(".");
                let values = data;

                for (const col of splittedColumn) {
                    if (values === null || values === undefined) {
                        return null;
                    }

                    if (Array.isArray(values)) {
                        const anidateValues = values.map((element) => {
                            return this.searchPropertyValues(element, col);
                        })
                        const [getArrayValue] = anidateValues[0];
                        objectValues.push(getArrayValue);
                        return null;
                    }

                    values = values[col];
                }

                objectValues.push(values);
            }
        }

        getValues(data, column);

        return objectValues;
    }

    private stringifyRecords(rows: any[], locale: string, visibleHeaders?: string[]) {
        var rowsString = '';

        for (let row of rows) {
            var rowString = '';

            for (let header of this.headers) {

                let cellValue = "";
                let arrayCellValue = [];

                if (header.id.includes(".")) {

                    arrayCellValue = this.searchPropertyValues(row, header.id);

                    const [value] = arrayCellValue;

                    cellValue = value;

                } else {
                    cellValue = row[header.id];
                }

                if (visibleHeaders === undefined || (visibleHeaders !== undefined && visibleHeaders.includes(header.id))) {
                    if (header.transformFunc !== undefined) {
                        rowString += header.transformFunc(cellValue) + FIELD_DELIMITER;
                    } else {
                        rowString += this.parseField(cellValue, locale);
                    }
                }
            }

            rowsString += rowString + "\n";
        }
        return rowsString;
    }

    private parseField(field: any, locale: string) {
        let parsedField = "";
        const localeParsed = locale.replace("_", "-");

        if (field !== undefined && field !== null) {
            switch (typeof (field)) {
                case "number":
                    if (isFloat(field)) {
                        parsedField += field.toLocaleString(localeParsed, { minimumFractionDigits: 10 }) + FIELD_DELIMITER;
                    } else {
                        parsedField += field.toString() + FIELD_DELIMITER;
                    }
                    break;
                case "object":
                    if (field.constructor.name === "Date") {
                        var options = { year: 'numeric', month: 'numeric', day: 'numeric' };
                        parsedField += field.toLocaleString(localeParsed, options) + FIELD_DELIMITER;
                    }
                    break;
                default:
                    parsedField += '"' + field.toString() + '"' + FIELD_DELIMITER;
                    break;
            }

        } else {
            parsedField += FIELD_DELIMITER;
        }
        return parsedField;
    }

    generateCsv(rows: any[], visibleHeaders?: string[], excludeBOM?: boolean) {
        var csv = ""
        if (!excludeBOM) {
            // Include the BOM for UTF8 (needed for excel)
            csv += "\uFEFF";
        }
        csv += this.getHeaders(visibleHeaders);
        csv += this.stringifyRecords(rows, this.locale, visibleHeaders);

        return csv;
    }
}

module.exports = CSVGenerator;